/*
 *        spInputPlugin.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spInputPluginP.h>
#include <sp/spRwPluginP.h>

spBool spIsInputPlugin(spPlugin *plugin)
{
    if (plugin == NULL) return SP_FALSE;

    if (spEqPluginType(plugin, SP_PLUGIN_INPUT) == SP_TRUE) {
        spDebug(80, "spIsInputPlugin", "SP_TRUE\n");
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

long _spReadPlugin(spPlugin *plugin, char *data, long length)
{
    long (*pread)(void *, char *, long) = NULL;

    spDebug(80, "_spReadPlugin", "length = %ld\n", length);
    
    if (spIsRwPlugin(plugin) == SP_TRUE) {
        pread = ((spRwPluginRec *)plugin->rec)->read;
    } else {
        if (spIsInputPlugin(plugin) == SP_FALSE) return -1;
    
        pread = ((spInputPluginRec *)plugin->rec)->read;
    }

    if (pread == NULL) return -1;
    if (plugin->instance == NULL
        && spInitPluginInstance(plugin) == SP_FALSE) return -1;
    
    return pread(plugin->instance, data, length);
}

long spReadPlugin(spPlugin *plugin, void *data, long length)
{
    return _spReadPlugin(plugin, (char *)data, length);
}

long _spReadPluginInByte(spPlugin *plugin, char *data, long data_size)
{
    int samp_byte;
    long length;
    long nread;

    spDebug(80, "_spReadPluginInByte", "data_size = %ld\n", data_size);
    
    if (spGetPluginDataSampleByte(plugin, &samp_byte) == SP_FALSE) {
        return -1;
    }

    length = data_size / (long)samp_byte;
    
    nread = _spReadPlugin(plugin, data, length);
    if (nread > 0) {
        return nread * (long)samp_byte;
    } else {
        return nread;
    }
}

long spReadPluginInByte(spPlugin *plugin, void *data, long data_size)
{
    return _spReadPluginInByte(plugin, (char *)data, data_size);
}

static long readPluginShiftWeighted(spPlugin *plugin, long prev_length, long shift_length,
                                    char *data, long length, spBool double_flag, double weight, spBool normalize_flag)
{
    long k;
    int samp_bit;
    int input_samp_bit;
    int file_samp_bit;
    long shift_byte;
    long read_length;
    long offset_length, offset_byte;
    spBool long_flag;
    long (*pread)(void *, char *, long) = NULL;

    spDebug(80, "readPluginShift", "prev_length = %ld, shift_length = %ld, length = %ld, double_flag = %d\n",
            prev_length, shift_length, length, double_flag);
    
    if (shift_length < 0 || data == NULL || length < 0) return -1;

    if (spIsRwPlugin(plugin) == SP_TRUE) {
        pread = ((spRwPluginRec *)plugin->rec)->read;
    } else {
        if (spIsInputPlugin(plugin) == SP_FALSE) {
            return -1;
        }
        pread = ((spInputPluginRec *)plugin->rec)->read;
    }

    if (pread == NULL) return -1;
    if (plugin->instance == NULL
        && spInitPluginInstance(plugin) == SP_FALSE) return -1;

    if (spGetPluginSampleBit(plugin, &file_samp_bit) == SP_FALSE) {
        return -1;
    }
    if (spGetPluginDataSampleBit(plugin, &input_samp_bit) == SP_FALSE) {
        return -1;
    }
    if (file_samp_bit > 16 && file_samp_bit <= 32) {
        long_flag = SP_TRUE;
    } else {
        long_flag = SP_FALSE;
    }
    if (double_flag == SP_TRUE) {
        samp_bit = sizeof(double) * 8;
    } else {
        samp_bit = input_samp_bit;
    }
    if (normalize_flag == SP_TRUE) {
        double factor;
        factor = spGetBitNormalizeFactor(file_samp_bit);
        spDebug(50, "readPluginShift", "factor = %f, original weight = %f\n", factor, weight);
        weight = weight / factor;
    }
    spDebug(80, "readPluginShift", "file_samp_bit = %d, input_samp_bit = %d, weight = %f\n",
            file_samp_bit, input_samp_bit, weight);

    if (prev_length <= 0) {
        offset_length = -prev_length;
    } else {
        offset_length = prev_length - shift_length;
    }
    read_length = length - MAX(offset_length, 0);
    shift_byte = shift_length * (samp_bit / 8);
    offset_byte = offset_length * (samp_bit / 8);
    
    if (prev_length < 0) {
        if (file_samp_bit > 33 || (file_samp_bit == 33 && input_samp_bit > 32)
            || double_flag == SP_TRUE) {
            double *ddata;
            
            ddata = (double *)data;
            for (k = 0; k < offset_length; k++) {
                ddata[k] = 0.0;
            }
        } else if (file_samp_bit == 33) {
            float *fdata;
            
            fdata = (float *)data;
            for (k = 0; k < offset_length; k++) {
                fdata[k] = 0.0;
            }
        } else {
            memset(data, 0, offset_byte);
        }
        
        if (read_length < 0) {
            read_length = 0;
        }
    } else {
        if (read_length < 0) return -1;
            
        if (offset_length > 0) {
            /* shift data */
            memmove(data, data + shift_byte, offset_byte);
        } else {
            offset_length = 0;
            offset_byte = 0;
        }
    }

    if (read_length > 0) {
        if ((read_length = pread(plugin->instance, data + offset_byte, read_length)) > 0) {
            spDebug(80, "readPluginShift", "read_length = %ld\n", read_length);
            if (double_flag == SP_TRUE && (input_samp_bit < sizeof(double) * 8 || long_flag == SP_TRUE)) {
                double *odata;
                
                odata = (double *)(data + offset_byte);
                
                if (file_samp_bit <= 16) {
                    short *sdata;
                
                    sdata = (short *)(data + offset_byte);
                    for (k = read_length - 1; k >= 0; k--) {
                        odata[k] = weight * (double)sdata[k];
                    }
                } else if (file_samp_bit <= 32) {
                    long *ldata;
                    
                    ldata = (long *)(data + offset_byte);
                    for (k = read_length - 1; k >= 0; k--) {
                        spDebug(100, "readPluginShift", "ldata[%ld] = %ld\n", k, ldata[k]);
                        odata[k] = weight * (double)ldata[k];
                    }
                } else if (file_samp_bit == 33 && input_samp_bit <= 32) {
                    float *fdata;

                    fdata = (float *)(data + offset_byte);
                    for (k = read_length - 1; k >= 0; k--) {
                        odata[k] = weight * (double)fdata[k];
                    }
                } else if (weight != 1.0) {
                    for (k = 0; k < read_length; k++) {
                        odata[k] *= weight;
                    }
                }
            }
        }
    }
    spDebug(80, "readPluginShift", "offset_length = %ld, read_length = %ld\n",
            offset_length, read_length);

    return read_length + offset_length;
}

long spReadPluginDoubleWeighted(spPlugin *plugin, double *data, long length, double weight)
{
    return readPluginShiftWeighted(plugin, 0, 0, (char *)data, length, SP_TRUE, weight, SP_TRUE);
}

long spReadPluginDouble(spPlugin *plugin, double *data, long length)
{
    return readPluginShiftWeighted(plugin, 0, 0, (char *)data, length, SP_TRUE, 1.0, SP_TRUE);
}

long _spReadPluginShift(spPlugin *plugin, long prev_length, long shift_length, char *data, long length)
{
    return readPluginShiftWeighted(plugin, prev_length, shift_length, data, length, SP_FALSE, 1.0, SP_FALSE);
}

long spReadPluginShift(spPlugin *plugin, long prev_length, long shift_length, void *data, long length)
{
    return _spReadPluginShift(plugin, prev_length, shift_length, (char *)data, length);
}

long spReadPluginShiftDoubleWeighted(spPlugin *plugin, long prev_length, long shift_length,
                                     double *data, long length, double weight)
{
    return readPluginShiftWeighted(plugin, prev_length, shift_length, (char *)data, length, SP_TRUE, weight, SP_TRUE);
}

long spReadPluginShiftDouble(spPlugin *plugin, long prev_length, long shift_length,
                             double *data, long length)
{
    return readPluginShiftWeighted(plugin, prev_length, shift_length, (char *)data, length, SP_TRUE, 1.0, SP_TRUE);
}

spBool spSeekPlugin(spPlugin *plugin, spLong pos)
{
    spBool (*pseek)(void *, spLong) = NULL;

    if (spIsRwPlugin(plugin) == SP_TRUE) {
        pseek = ((spRwPluginRec *)plugin->rec)->seek;
    } else {
        if (spIsInputPlugin(plugin) == SP_FALSE) return SP_FALSE;
    
        pseek = ((spInputPluginRec *)plugin->rec)->seek;
    }

    if (pseek == NULL) return SP_FALSE;
    if (plugin->instance == NULL
        && spInitPluginInstance(plugin) == SP_FALSE) return SP_FALSE;
    
    return pseek(plugin->instance, pos);
}

spLong spGetPluginTotalLength(spPlugin *plugin)
{
    spLong (*get_total_length)(void *) = NULL;
    
    if (spIsRwPlugin(plugin) == SP_TRUE) {
        get_total_length = ((spRwPluginRec *)plugin->rec)->get_total_length;
    } else {
        if (spIsInputPlugin(plugin) == SP_FALSE) return -1;
    
        get_total_length = ((spInputPluginRec *)plugin->rec)->get_total_length;
    }

    if (get_total_length == NULL) return -1;
    if (plugin->instance == NULL
        && spInitPluginInstance(plugin) == SP_FALSE) return -1;
    
    return get_total_length(plugin->instance);
}
